(* Copy files using Windows Shell
   ==============================
   Source: local drive, net resource or portable device (phone) +
   Destination: local drive

    Dr. J. Rathlev, D-24222 Schwentinental

   The contents of this file may be used under the terms of the
   Mozilla Public License ("MPL") or
   GNU Lesser General Public License Version 2 or later (the "LGPL")

   Software distributed under this License is distributed on an "AS IS" basis,
   WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
   the specific language governing rights and limitations under the License.

   J. Rathlev, June 2022
   last modified: June 2022
   *)

unit CopyFilesMain;

interface

uses
  Winapi.Windows, Winapi.Messages, System.SysUtils, System.Variants, System.Classes, Vcl.Graphics,
  Vcl.Controls, Vcl.Forms, Vcl.Dialogs, Vcl.ComCtrls, Vcl.Shell.ShellCtrls,
  Vcl.StdCtrls, Vcl.Buttons, Vcl.ExtCtrls;

type
  TMainForm = class(TForm)
    ShellTreeView: TShellTreeView;
    ShellListView: TShellListView;
    sbDestDir: TSpeedButton;
    bbCopy: TBitBtn;
    edDestDir: TLabeledEdit;
    procedure FormCreate(Sender: TObject);
    procedure sbDestDirClick(Sender: TObject);
    procedure bbCopyClick(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure ShellTreeViewClick(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure FormResize(Sender: TObject);
    procedure ShellTreeViewDblClick(Sender: TObject);
  private
    { Private-Deklarationen }
    IniName,
    PersDir,
    AppData,
    LastPath : string;
  public
    { Public-Deklarationen }
  end;

var
  MainForm: TMainForm;

implementation

{$R *.dfm}

uses Winapi.ShlObj, Winapi.ActiveX, System.Win.ComObj, Winapi.ShellApi,
  System.IniFiles, Vcl.FileCtrl, WinShell;

const
  IniExt = '.ini';

  (* INI-Sections *)
  CfGSekt = 'Config';

  (* INI-Variables *)
  IniPath = 'LastPath';
  IniDest = 'DestPath';

procedure TMainForm.FormCreate(Sender: TObject);
begin
  PersDir:=GetPersonalFolder;
  AppData:=GetAppDataFolder;
  IniName:=IncludeTrailingPathDelimiter(AppData)+ChangeFileExt(ExtractFileName(Application.ExeName),IniExt);
  with TIniFile.Create(IniName) do begin
    LastPath:=ReadString(CfgSekt,IniPath,PersDir);
    edDestDir.Text:=ReadString(CfgSekt,IniDest,PersDir);
    Free;
    end;
  end;

procedure TMainForm.FormDestroy(Sender: TObject);
begin
  with TIniFile.Create(IniName) do begin
    WriteString(CfgSekt,IniPath,LastPath);
    WriteString(CfgSekt,IniDest,edDestDir.Text);
    Free;
    end;
  end;

procedure TMainForm.FormResize(Sender: TObject);
begin
  with ShellListView do SetColWidths([GetWidth-48,12,15,17]) // Anzahl Zeichen pro Spalte
  end;

procedure TMainForm.FormShow(Sender: TObject);
begin
  try
    ShellTreeView.Path:=LastPath;
  except
    ShellTreeView.Path:=PersDir;
    end;
  end;

procedure TMainForm.sbDestDirClick(Sender: TObject);
var
  s : string;
  sf : TShellFolder;
begin
  s:=edDestDir.Text;
  if length(s)=0 then begin
    sf:=ShellTreeView.SelectedFolder;
    if assigned(sf) then s:=sf.PathName;
    end;
  if length(s)=0 then s:=GetPersonalFolder;
  if SelectDirectory(s,[sdAllowCreate,sdPerformCreate,sdPrompt],0) then begin
//  if ShellDirDialog.Execute ('Destination folder',true,true,true,s,s) then begin
    edDestDir.Text:=s;
    end;
  end;

function GetPortableParseName (AbsoluteID: PItemIDLIst) : string;
var
  pn : PWideChar;
  sa,se : string;
  n1,n2 : integer;
begin
  OleCheck(SHGetNameFromIDList(AbsoluteID,SIGDN_DESKTOPABSOLUTEPARSING,pn));
  sa:=pn; CoTaskMemFree(pn); pn:=nil;
  OleCheck(SHGetNameFromIDList(AbsoluteID,SIGDN_DESKTOPABSOLUTEEDITING,pn));
  se:=pn; CoTaskMemFree(pn); pn:=nil;
  n1:=pos('\\\?\',sa);
  n1:=pos('\',sa,n1+5);
  if n1>0 then begin
    n2:=pos('\',se);
    n2:=pos('\',se,n2+1);
    Result:=copy(sa,1,n1)+copy(se,n2+1,length(se));
    end
  else Result:=sa;
  end;

procedure TMainForm.ShellTreeViewClick(Sender: TObject);
begin
  with ShellTreeView.SelectedFolder do begin
    if (fpFileSystem in Properties) then LastPath:=PathName
    else LastPath:=GetPortableParseName(AbsoluteID);
    end;
  end;

procedure TMainForm.ShellTreeViewDblClick(Sender: TObject);
begin
  with ShellTreeView do if assigned(Selected) then Refresh(Selected);
  end;

procedure TMainForm.bbCopyClick(Sender: TObject);
var
  i,n,j : integer;
  fileOp : IFileOperation;
  siSrcList : IShellItemArray;
  idList : array of PItemIDList;
  siSrcFile,siDestFolder : IShellItem;
begin
  with ShellListView do if assigned(Selected) then begin
    n:=Selected.Index;
    OleCheck(CoCreateInstance(CLSID_FileOperation,nil,CLSCTX_ALL,IFileOperation,fileOp));
    if SelCount=1 then with SelectedFolder do begin
      OleCheck(SHCreateItemFromIDList(AbsoluteID,IShellItem,siSrcFile));
      OleCheck(SHCreateItemFromParsingName(PChar(edDestDir.Text),nil,IShellItem,siDestFolder));
      OleCheck(fileOp.CopyItem(siSrcFile,siDestFolder,pchar(DisplayName),nil));
      OleCheck(fileOp.PerformOperations);
      end
    else begin
      SetLength(idList,SelCount); j:=0;
      OleCheck(fileOp.SetOperationFlags(FOF_FILESONLY+FOF_NOCONFIRMMKDIR+FOF_NO_CONNECTED_ELEMENTS));
      for i:=n to Items.Count-1 do if Items[i].Selected then begin
        idList[j]:=Folders[i].AbsoluteID; inc(j);
        end;
      OleCheck(SHCreateShellItemArrayFromIDLists(SelCount,@idList[0],siSrcList));
      OleCheck(SHCreateItemFromParsingName(PChar(edDestDir.Text),nil,IShellItem,siDestFolder));
      OleCheck(fileOp.CopyItems(siSrcList,siDestFolder));
      OleCheck(fileOp.PerformOperations);
      idList:=nil;
      end;
    end;
  end;

end.
